{-
By Lewis Crawford (s1203531

See README.txt for more details

A 3D fractal renderer that can draw either the mandelbox, the serpinski gasket, or a set of spheres.
It uses a distance estimation formula, and a raymarching algorithm to build up an image on a per-pixel basis.
There is also an option to enable a more sophisticated colouring algorithm using phong-reflection techniques and
a list of configurable light sources. This is a lot slower, and the detail is not as good when doing close-ups of the Mandelbox
(but it makes spheres look much nicer).

All the parameters in below can be tweaked to get different images. 

-}




import Graphics.Rendering.OpenGL
import Graphics.UI.GLUT
import Data.Int
import Control.Parallel.Strategies hiding (dot)

-------------------------------------------------------------------------------------------
----General Rendering Paramaters ----------------------------------------------------------
-------------------------------------------------------------------------------------------


--Suggested values:

--To see the whole scale 2 mandelbox around a distance of 20, maxRaySteps of 50, and minimumDistance of 0.005 seems to work nicely
--Yrot -1.1, centre of focus (1,-1,0), and distance 18 is also a good angle

--A 100x zoom on the surface of the box.
--Mandelbox with 58 raysteps, 0.0000725 minumum distance, centreOfFocus (5,5,0), distanceFromFocus = 15, yRot = 0.5 (others 0), and a 100x zoom (others set to defaults).

--Focusing on point (6,6,0) with a min distance of 0.001, complex lighting off, and a yRot of 0.5 is another interesting shot.


--A good combination for the serpinski gasket of iteration 15 is a distance of 3.75 from the origin, a maxRaySteps of 20, a min distance of 0.005,
--and xRot = -0.1, yRot = 0.5, zRot = 0.5 (all others at defaults).

data RenderObject =  Mandelbox | SerpinskiGasket | Spheres
  deriving (Eq)
--Determines what this renderer draws. The most interesting is the mandelbox, but the serpinski gasket looks quite good as well.
--The spheres option was the first thing I got this code working with, and has been added for variety. Each type has it's own set
--of unique parameters that determine how they are drawn (see bellow the general rendering paramaters section).
objectToRender = Mandelbox

--The height of the screen in pixels. Determines resolution of final render.
--Higher resoultions take much longer to render. For the mandelbox on my laptop
--anything above 300px takes a non-trivial ammount of time to render. Other objects are more lenient,
--but for the mandelbox try 200px renders to tweak settings before upping the resolution.
screenWidth = 300 
--The height of the screen in pixels.
screenHeight = 300

--Maxumum number of steps raymarcher should take before determining whether a point is near enough to the object to render.
--If this number is reached before the point on the ray is within the minumum distance, the point is considered not
--do be part of the image.
maxRaySteps = 50
--The minimum distance to the object that the raymarcher should define as being close enought to the object.
--If this is too small, and the max ray steps in not large enough, the image will appear grainy, and have chunks missing.
minimumDistance = 0.005

--The colour that the object should be at its lightest point. Darker versions of this colour are used for shadows.
foregroundColour = (1,1,1)

--The colour for the background (i.e. empty space around the object) should be. A very slight grey is used to emphasise the
--object being rendered.
backgroundColour = Color3 0.04 0.04 0.04

--The point in space that the centre of the camera should focus on. All subsequent camera transformations use
--this point as the origin. The units are those of the world you are rendering, not pixels or screen percent.
--Usually (0,0,0)
centreOfFocus = (0,0,0)

--The number of world-units that the camera is positioned away from the object it is viewing. Initially,
--before any rotations, this distance is taken to be along the negative z axis.
distanceFromFocus = 20

--Allow the camera to zoom in on a point. If scale factor is 0.1, it will result in a 10x zoom, and if it is 2 it will result in
--zooming out twice as far. Allows you to zoom in on a fixed point without moving the camera. Normally this will be 1.
scaleFactor = 1

--WARNING these values may be unpredictable, as the rotation is not fully tested (it was a last minute feature)

--The rotation in radians that the camera has been moved around the x axis with the origin at its centre of vision (i.e. centreOfFocus)
xRot = 0
--The rotation around the y axis in radians
yRot = 0
--The rotation around the y axis in radians
zRot = 0

-----------------------Lighting Parameters----------------------------------------------------------------


--If this is True, then the more complex phong reflection style of lighting will be used.
--If False, a simpler but faster shading algorithm that does not need to calculate the normal
--(saving 6 extra repetitions of the DE formula per ray)
useComplexLighting = False

--Light constants:
iA = 0.2 --Default ambient light

--The type for a light source - position, specular component, diffuse component
type LightSource = (Vec3,Float,Float)

--The list of light sources the object is illuminated by.
lightSources :: [LightSource]
lightSources = [((11,11,-12),0.5,0.5)]

--Surface constants:
kS = 0.5 --Spectral reflection constanct for the surface of the rendered object.
kD = 0.5 --Diffuse reflection constant for the surface of the rendered object.
kA = 0.2 --Ambient reflection component for the surface of the rendered object.
shininess = 10 --How mirror-like the surface is.


-----------------------Mandelbox Parameters-------------------------------------------------------------

--The number of loops the distance estimation forumula should repeat its generation process. The higher this is,
--the more detail the box can have. Around 10 seems good.
mandelboxIterations = 10
--A constant that determines how the box is drawn. At 2 the box seems pretty continuous, but at 3 it start to break up a bit.
mandelboxScale = 2
--Default value of 1
fixedRadius = 1.0
--Default value of 0.5
minRadius = 0.5

-----------------------Serpinski Gasket Parameters------------------------------------------------------

--The number of loops the distance estimation forumula should repeat its generation process. The higher this is,
--the more subdivisions the tetrahedron will have. Around 15 seems good.
serpinskiIterations = 16
--How much to scale by each iteration of the DE loop. Usually 2.
serpinskiScale = 2
--The point that space is folded around in the distance estimation formula. Changing
--this makes some interesting looking folded tetrahedrons. Usually (1,1,1) to get a recular gasket.
serpinskiOffset = (1,1,1)


---------------------Spheres Paramaters-----------------------------------------------------------------

--A sphere has a position in the 3D space, and a radius, stored in a tuple.
type Sphere = (Vec3,Float)

--This list of spheres are what will be rendered in on screen with the distance estiation formula.
spheres :: [Sphere]
spheres = [((0,0,0),0.2) , ((0,1,1),0.5) , ((2,1,0.5),1)]


----------------------------------------------------------------------------------------------
--------------------------------------The Code------------------------------------------------
----------------------------------------------------------------------------------------------

--The main method
--Sets up the screen for drawing on, and hands over to the mainloop to handle the drawing process etc.
main :: IO ()
main = do
  (progname, _) <- getArgsAndInitialize
  initialDisplayMode $= [DoubleBuffered, RGBMode, WithDepthBuffer]
  createWindow "Mandelbox"
  windowSize $= Size (fromIntegral screenWidth) (fromIntegral screenHeight)
  displayCallback $= display
  mainLoop

  
--Draws the list of pixels and colour values on to the screen in a grid
--As each pixel is drawn as a point, enlarging the screen will gause gaps to appear in the picture.
display :: IO ()
display = do
	clear [ColorBuffer, DepthBuffer]
	renderPrimitive Points (foldr1 (>>) commands)
	swapBuffers
	where
		commands = parMap rseq (\(c,v) -> color c >> vertex v) pixels
	
	
--The list of pixels that make up the picture. Each consists of an openGL rgb colour value, and an openGL Vertex, which contains its
--coordinates on the screen (the screen's coordinates go from -1 to 1, as the openGL built in camera is never moved).
pixels :: [(Color3 GLfloat, Vertex3 GLfloat)]
pixels = getPixels objectToRender centreOfFocus distanceFromFocus xRot yRot zRot scaleFactor screenWidth screenHeight foregroundColour


--Gets the colour values of all the pixels that need to be rendered on the screen.
--First it calculates the projected size of the camera's 2D line of sight towards the given centre of focus (after rotation etc.)
--It then calculates a unit vector in the direction from the camera's centre to each point on the projected 2D screen.
--After calculating one ray for every pixel, it then uses the traceRay function to evaluate its colour.
--This method returns a list 
getPixels :: RenderObject -> Vec3 -> Float -> Float -> Float -> Float -> Float -> Int -> Int -> (Float,Float,Float) -> [(Color3 GLfloat, Vertex3 GLfloat)]
getPixels objectToRender centreOfFocus zDist xRot yRot zRot scaleFactor widthPx heightPx colour = [(getColourForXY (fromIntegral x) (fromIntegral y), Vertex3 ((fromIntegral x)/(fromIntegral w2)) ((fromIntegral y)/(fromIntegral h2)) 0)| x <- widthIndexes, y <- heightIndexes]
	where
		relPos = (rotateVec xRot yRot zRot (0,0,(-zDist)))
		cannonicalPos = relPos `addVec` centreOfFocus
		heightDir = (rotateVec xRot yRot zRot (0,1,0))
		widthDir = (rotateVec xRot yRot zRot (1,0,0))
		h2 = heightPx `div` 2
		heightIndexes = [(-h2)..h2]
		w2 = widthPx `div` 2
		widthIndexes = [(-w2)..w2]
		widthStepSize =  zDist*scaleFactor/(fromIntegral widthPx)
		heightStepSize = zDist*scaleFactor/(fromIntegral heightPx)
		getRay x y = normalizeVec ((((y*heightStepSize) `scalarMult` heightDir) `addVec` ((x*widthStepSize) `scalarMult` widthDir)) `minusVec` relPos)
		getColourForXY x y = traceRay cannonicalPos (getRay x y) colour objectToRender


--Uses a ray marching technique to calculate if a ray from the given start point comes within a minimum
--distance of the object being rendered. It recursively tests roughly how far away the current
--point is from the nearest point on the object (using the distance estimation formula), and then marches that
--distance along the ray. When it reaches the new point, it checks how far away it is, and if it is
--withing the minimum bound, it stops, as it is close enough to the object to calculate a colour value.
--If using complex lighting, it calculates the normal to that point, and illuminates it from light sources.
--If using simple lighting, it calculates a colour value based on the number of ray steps it needed to reach the point.
traceRay  :: Vec3 -> Vec3 -> (Float, Float, Float) -> RenderObject -> Color3 GLfloat
traceRay  from direction colour@(r,g,b) objectToRender | (objectToRender == Mandelbox) = f 0 0 mandelboxDE
											   | (objectToRender == SerpinskiGasket) = f 0 0 serpinskiDE
											   | (objectToRender == Spheres) = f 0 0 spheresDE
											   | otherwise = error "Invalid render object type."
	where
		f :: Float -> Int -> (Vec3 -> Float) -> Color3 GLfloat
		f totalDistance steps distanceEstimator | steps > maxRaySteps = backgroundColour
												| distance < minimumDistance = if useComplexLighting 
																				then illuminate from p n lightSources colour
																				else Color3 (r*c) (g*c) (b*c)
												| otherwise = f (totalDistance + distance) (steps+1) distanceEstimator
			where
				distance = distanceEstimator p
				p = from `addVec` (totalDistance `scalarMult` direction)
				n = getNormal p (1,0,0) (0,1,0) (0,0,1) distanceEstimator
				c = 1 - (fromIntegral(steps)/fromIntegral(maxRaySteps))
				
--Uses the phong reflection techinque to calculate the brightness of the light shining on each point.
--Finds the total illumination from a list of point sources on a particular material.
illuminate :: Vec3 -> Vec3 -> Vec3 -> [LightSource] -> (Float,Float,Float) -> Color3 GLfloat
illuminate cameraPos p n lights (r,g,b) = Color3 (illum r) (illum g) (illum b)
	where
		illum x = min 1 (x*lightIntensity)
		lightIntensity = min 1 (kA * iA + sum [illumFromSource ls | ls <- lightSources])
		illumFromSource (lightPos, iS, iD) = (kD*(l `dot` n)*iD) + (kS*((r `dot` v)**shininess)*iS)
			where
				l = normalizeVec (lightPos `minusVec` p)
				r = normalizeVec (cameraPos `minusVec` p)
				v = normalizeVec (((2 * (l `dot` n)) `scalarMult` n) `minusVec` l)
				
				
--------------------------------------------------------------------------------------------------------			
----------------------------------Distance Estimation Formulas------------------------------------------
--------------------------------------------------------------------------------------------------------
			
--Use this to draw the mandelbox.
mandelboxDE :: Vec3 -> Float
mandelboxDE p = f 0 p mandelboxScale
	where 
		f :: Int -> Vec3 -> Float -> Float
		f n v deFactor | n >= mandelboxIterations = (lengthVec v)/(abs deFactor)
					   | otherwise = f (n+1) ((mandelboxScale `scalarMult` (fst changedVals)) `addVec` p) ((snd changedVals)*mandelboxScale)
			where
				changedVals = change v deFactor
				change v deFactor = g2 (g1 v) deFactor
					where 
						g1 v@(x,y,z) = (alter x, alter y, alter z)
						alter q | q > 1 = 2-q
								| q < -1 = -2-q
								| otherwise = q
						g2 v@(x,y,z) deFactor | r2 < mR2 = ((a x,a y,a z), a deFactor)
											  | r2 < fR2 = ((b x,b y,b z), b deFactor)
											  | otherwise = (v,deFactor)
							where
								a c = c*fR2/mR2
								b c = c*fR2/r2
								r2 = x*x + y*y + z*z
		fR2 = fixedRadius * fixedRadius
		mR2 = minRadius * minRadius

--Use this to draw the serpinski gasket
serpinskiDE :: Vec3 -> Float
serpinskiDE p = f 0 p
	where 
		f :: Int -> Vec3 -> Float
		f n z | n >= serpinskiIterations = (lengthVec z) * (serpinskiScale^^(-n))
					  | otherwise = f (n+1) ((serpinskiScale `scalarMult` z') `minusVec` ((serpinskiScale-1) `scalarMult` serpinskiOffset))
			where
				z' = (fold3.fold2.fold1) z
				fold1 v@(x,y,z) | (x + y) < 0 = (-y,-x,z)
								| otherwise = v
				fold2 v@(x,y,z) | (x + z) < 0 = (-z,y,-x)
								| otherwise = v
				fold3 v@(x,y,z) | (y + z) < 0 = (x,-z,-y)
								| otherwise = v

								
--Finds the minimum distance from any of the spheres in the list. This
--process of finding the minimum using a list of DE formulas can be used for
--any image, enabling you to build complex scenes (if you have time).
spheresDE p = foldr1 min (map (\(v,r) -> (lengthVec (p `minusVec` v)) - r) spheres)

--------------------------------------------------------------------------------------------------------
-----------------------------------Vector operations----------------------------------------------------
--------------------------------------------------------------------------------------------------------

--The 3D vector type - a triple of floating point coordinate (x,y,z)
type Vec3 = (Float, Float, Float)
			  
--Vector addition
addVec :: Vec3 -> Vec3 -> Vec3
addVec (x,y,z) (a,b,c) = (x+a,y+b,z+c)

--Vector subtraction
minusVec :: Vec3 -> Vec3 -> Vec3
minusVec (x,y,z) (a,b,c) = (x-a,y-b,z-c)

--Multipliation of a vector by a floating point scalar
scalarMult :: Float -> Vec3 -> Vec3
scalarMult c (x,y,z) = (c*x, c*y, c*z)

--The dot product of two 3D vectors
dot :: Vec3 -> Vec3 -> Float
dot (x,y,z) (a,b,c) = x*a + y*b + z*c

--The length of a 3D vector
lengthVec :: Vec3 -> Float
lengthVec (x,y,z) = sqrt (x*x + y*y + z*z)

--Returns a unit-length vector in the same direction as it's parameter
normalizeVec :: Vec3 -> Vec3
normalizeVec v = (1 /(lengthVec v)) `scalarMult` v

--Rotates a vector round the x, y, and z directions about the origin.
--Derived from the 3D rotation matrix, but without the need to explicitly code
--matrix types and transformations. All angles are in radians.
rotateVec :: Float -> Float -> Float -> Vec3 -> Vec3
rotateVec phi theta psi (x,y,z) = (x' , y' , z')--xrot, yrot, zrot (all in radians)
	where
		x' = x1' + x2' + x3'
		x1' = x * (cos theta) * (cos psi)
		x2' = y * ((-(cos phi) * (sin psi)) + ((sin phi)*(sin theta)*(cos psi)))
		x3' = z * (((sin phi)*(sin psi))+((cos phi)*(sin theta)*(cos psi)))
		y' = y1' + y2' + x3'
		y1' = x * (cos theta) * (sin psi)
		y2' = y * ((cos phi)*(cos psi) + (sin phi)*(sin theta)*(sin psi))
		y3' = z * ((-(sin phi))*(cos psi) + (cos phi)*(sin theta)*(sin psi))
		z' = z1' + z2' + z3'
		z1' = x * (-(sin theta))
		z2' = y * (sin phi)*(cos theta)
		z3' = z * (cos phi) * (cos theta)
		
--Get the normal to the 3D object at the given point using the distance estimation formula
--Normals are used for lighting effects. They are calculated using the finite difference method.
getNormal :: Vec3 -> Vec3 -> Vec3 -> Vec3 -> (Vec3 -> Float) -> Vec3
getNormal pos xDir yDir zDir de = normalizeVec (nx,ny,nz)
	where
		nx = de(pos `addVec` xDir)-de(pos `minusVec` xDir)
		ny = de(pos `addVec` yDir)-de(pos `minusVec` yDir)
		nz = de(pos `addVec` zDir)-de(pos `minusVec` zDir)
